#include "mbed.h"
#include "ble/BLE.h"
#include "BatteryService.h"
#include "DeviceInformationService.h"

#define BLE_UUID_DESCRIPTOR_REPORT_REFERENCE 0x2908

#ifndef NDEBUG
#define HID_DEBUG(...) printf(__VA_ARGS__)
#else
#define HID_DEBUG(...)
#endif

InterruptIn bt = InterruptIn(p28);
//const static uint8_t AdvData[] = {"advert"};

static uint8_t consumerInputReportData[] = {0x10};
static const uint8_t consumerEmptyInputReportData[] = {0};

static uint8_t CONSUMER_REPORT_MAP[] = {
    0x05, 0x0c,                    // USAGE_PAGE (Consumer Devices)
    0x09, 0x01,                    // USAGE (Consumer Control)
    0xa1, 0x01,                    // COLLECTION (Application)
                                    // -------------------- common global items
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x75, 0x01,                    //   REPORT_SIZE (1)    - each field occupies 1 bit
                                   // -------------------- misc bits
    0x95, 0x05,                    //   REPORT_COUNT (5)
    0x09, 0xb5,                    //   USAGE (Scan Next Track)
    0x09, 0xb6,                    //   USAGE (Scan Previous Track)
    0x09, 0xb7,                    //   USAGE (Stop)
    0x09, 0xcd,                    //   USAGE (Play/Pause)
    0x09, 0xe2,                    //   USAGE (Mute)
    0x09, 0xb9,                    //   random play
    0x81, 0x06,                    //   INPUT (Data,Var,Rel)  - relative inputs
                                   // -------------------- volume up/down bits
    0x95, 0x02,                    //   REPORT_COUNT (2)
    0x09, 0xe9,                    //   USAGE (Volume Up)
    0x09, 0xea,                    //   USAGE (Volume Down)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)  - absolute inputs
                                   // -------------------- padding bit
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x81, 0x01,                    //   INPUT (Cnst,Ary,Abs)
    0xc0                           // END_COLLECTION
};

typedef struct {
    uint8_t ID;
    uint8_t type;
} report_reference_t;

bool isConnected;
static BLEProtocol::Address_t peerAddress;
report_reference_t inputReportReferenceData;
GattAttribute inputReportReferenceDescriptor(BLE_UUID_DESCRIPTOR_REPORT_REFERENCE,
            (uint8_t *)&inputReportReferenceData, 2, 2);

GattAttribute** inputReportDescriptors() {
    inputReportReferenceData.ID = 0;
    inputReportReferenceData.type = 0x1;

    static GattAttribute * descs[] = {
        &inputReportReferenceDescriptor,
    };
    return descs;
}

 GattCharacteristic inputReportCharacteristic(GattCharacteristic::UUID_REPORT_CHAR,
            (uint8_t *)consumerInputReportData, sizeof(consumerInputReportData), sizeof(consumerInputReportData),
              GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_READ
            | GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_NOTIFY
            | GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_WRITE,
            inputReportDescriptors(), 1);

 GattCharacteristic reportMapCharacteristic(GattCharacteristic::UUID_REPORT_MAP_CHAR,
            (uint8_t*) CONSUMER_REPORT_MAP, sizeof(CONSUMER_REPORT_MAP), sizeof(CONSUMER_REPORT_MAP),
            GattCharacteristic::BLE_GATT_CHAR_PROPERTIES_READ);       
    
    
    GattCharacteristic *characteristics[] = {
        &reportMapCharacteristic,
        &inputReportCharacteristic
    };

void securityManager_onSecuritySetupInitiated(Gap::Handle_t, bool allowBonding, bool requireMITM, SecurityManager::SecurityIOCapabilities_t iocaps)
{
    HID_DEBUG("securityManager_onSecuritySetupInitiated\r\n");
}

void printWhiteList(){
    Gap::Whitelist_t whiteList;
    BLE::Instance().securityManager().getAddressesFromBondTable(whiteList);
    
    HID_DEBUG("white list counter \%d \r\n", whiteList.size);

}



void securityManager_onSecuritySetupCompleted(Gap::Handle_t, SecurityManager::SecurityCompletionStatus_t status)
{
    HID_DEBUG("securityManager_onSecuritySetupCompleted -> Status: \%d\r\n", status);
    if (status == SecurityManager::SEC_STATUS_SUCCESS) {
        HID_DEBUG("Security success %d\r\n", status);
        HID_DEBUG("Set whitelist\r\n");
        Gap::Whitelist_t whitelist;
        whitelist.size = 1;
        whitelist.capacity = 1;
        whitelist.addresses = &peerAddress;

        BLE::Instance(BLE::DEFAULT_INSTANCE).gap().setWhitelist(whitelist);
         
    } else {
        HID_DEBUG("Security failed %d\r\n", status);
    }
}

void securityManager_onLinkSecured(Gap::Handle_t handle, SecurityManager::SecurityMode_t securityMode)
{
    HID_DEBUG("securityManager_onLinkSecured started -> securityMode: \%d\r\n", securityMode);
    HID_DEBUG("securityManager_onLinkSecured ended -> securityMode: \%d\r\n", securityMode);
}

void securityManager_onPasskeyDisplay(Gap::Handle_t handle, const SecurityManager::Passkey_t passkey){
    HID_DEBUG("securityManager_oonPasskeyDisplay\r\n");
}

void securityManager_onSecurityContextStored(Gap::Handle_t handle){
    HID_DEBUG("securityManager_onSecurityContextStored\r\n");
    HID_DEBUG("securityManager_onSecurityContextStoredEnded\r\n");
}

void ble_onTimeout(Gap::TimeoutSource_t source){
    HID_DEBUG("ble_onTimeout-> source \%d \r\n", source);
}

void gattServer_onUpdatesEnabled(GattAttribute::Handle_t handle){
    HID_DEBUG("gattServer_onUpdatesEnabled \%d \r\n",  handle);
    HID_DEBUG("gattServer_onUpdatesEnabled ended \%d \r\n",  handle);
}

void gattServer_onUpdatesDisabled(GattAttribute::Handle_t handle){
    HID_DEBUG("gattServer_onUpdatesDisabled \%d \r\n",  handle);
    HID_DEBUG("gattServer_onUpdatesDisabled ended \%d \r\n",  handle);
}

void gattServer_onConfirmationReceived(GattAttribute::Handle_t handle){
    HID_DEBUG("gattServer_onConfirmationReceived \%d \r\n",  handle);
    HID_DEBUG("gattServer_onConfirmationEnded \%d \r\n",  handle);
}



void ble_onDisconnected(const Gap::DisconnectionCallbackParams_t *params)
{
    HID_DEBUG("bled disconnected starting Advertising\r\n");
    isConnected=false;
    BLE::Instance().gap().startAdvertising();
    HID_DEBUG("bled disconnected ended\r\n");
    
}


void ble_onConnected(const Gap::ConnectionCallbackParams_t *params)
{
    HID_DEBUG("bled connected started\r\n");
    isConnected=true;
    
    peerAddress.type = params->peerAddrType;
    memcpy(peerAddress.address, params->peerAddr, Gap::ADDR_LEN);

    BLEProtocol::Address_t peerAddresses[2];
    Gap::Whitelist_t whitelist;
    whitelist.size = 0;
    whitelist.capacity = 2;
    whitelist.addresses = peerAddresses;
    BLE::Instance(BLE::DEFAULT_INSTANCE).gap().getWhitelist(whitelist);
    HID_DEBUG("getWhitelist %d\r\n", whitelist.size);
    for (int i = 0; i < whitelist.size; i++) {
    if (whitelist.addresses[i].type == params->peerAddrType &&
            memcmp(whitelist.addresses[i].address, params->peerAddr, Gap::ADDR_LEN) == 0) {
            HID_DEBUG("peer is found exiting onconnect\r\n");
            return;
        }
    }
    HID_DEBUG("peer is not found in whitelist exiting on connect\r\n");
}



void InitializeSecurity(BLE &b)
    {
        HID_DEBUG("[main..InitializeSecurity]started\r\n");
       
        b.securityManager().onSecuritySetupInitiated(securityManager_onSecuritySetupInitiated);
        b.securityManager().onSecuritySetupCompleted(securityManager_onSecuritySetupCompleted);
        b.securityManager().onLinkSecured(securityManager_onLinkSecured);
        b.securityManager().onSecurityContextStored(securityManager_onSecurityContextStored);
        b.securityManager().onPasskeyDisplay(securityManager_onPasskeyDisplay);
        b.initializeSecurity();
        HID_DEBUG("[main..InitializeSecurity] ended \r\n");
    } 

 void bleInitComplete(BLE::InitializationCompleteCallbackContext *params)
{
    HID_DEBUG("[main..bleInitComplete] started \r\n");
    BLE&        ble   = params->ble;
    
    ble.gap().onDisconnection(ble_onDisconnected);
    ble.gap().onConnection(ble_onConnected);
    ble.gap().onTimeout(ble_onTimeout);
    ble.gattServer().onUpdatesEnabled(gattServer_onUpdatesEnabled);
    ble.gattServer().onUpdatesDisabled(gattServer_onUpdatesDisabled);
    ble.gattServer().onUpdatesDisabled(gattServer_onConfirmationReceived);
    InitializeSecurity(ble);
   
    
    ble.gap().setDeviceName("Consumer service");
    
    
    uint16_t minConnectionInterval = Gap::MSEC_TO_GAP_DURATION_UNITS(7.5);
    uint16_t maxConnectionInterval = Gap::MSEC_TO_GAP_DURATION_UNITS(30);
    uint16_t slaveLatency = 6;
    uint16_t connectionSupervisionTimeout = Gap::MSEC_TO_GAP_DURATION_UNITS(430);
    Gap::ConnectionParams_t connectionParams = {
        minConnectionInterval,
        maxConnectionInterval,
        slaveLatency,
        connectionSupervisionTimeout
    };
    
    ble.gap().setPreferredConnectionParams(&connectionParams);
    
     DeviceInformationService deviceInfo(ble, "ARM", "m1", "abc", "def", "ghi", "jkl");
    BatteryService batteryInfo(ble, 80);
    GattService consumerService(GattService::UUID_HUMAN_INTERFACE_DEVICE_SERVICE,
                        characteristics, 2);                   
    ble.gattServer().addService(consumerService);
    
    
    ble.gap().accumulateAdvertisingPayload(GapAdvertisingData::BREDR_NOT_SUPPORTED | GapAdvertisingData::LE_GENERAL_DISCOVERABLE );
    ble.gap().setAdvertisingType(GapAdvertisingParams::ADV_CONNECTABLE_UNDIRECTED);
    
     
    static const uint16_t servicesIds_list[] =  {GattService::UUID_HUMAN_INTERFACE_DEVICE_SERVICE,
            GattService::UUID_DEVICE_INFORMATION_SERVICE,
            GattService::UUID_BATTERY_SERVICE};
    ble.gap().accumulateAdvertisingPayload(GapAdvertisingData::COMPLETE_LIST_16BIT_SERVICE_IDS,
               (uint8_t *)servicesIds_list, sizeof(servicesIds_list));
    
    ble.gap().accumulateAdvertisingPayload(GapAdvertisingData::KEYBOARD);
 
        
    ble.gap().setAdvertisingInterval(20); /* 100ms */
    ble.gap().startAdvertising();
    HID_DEBUG("[main..bleInitComplete] ended \r\n");
}


void button_pressed(){
    HID_DEBUG("button_pressed started");
    if(isConnected){
        ble_error_t result;
        result = BLE::Instance(BLE::DEFAULT_INSTANCE).updateCharacteristicValue(inputReportCharacteristic.getValueHandle(), (uint8_t*)consumerInputReportData, sizeof(consumerInputReportData));
        HID_DEBUG("result \%d \r\n", result);
        result = BLE::Instance(BLE::DEFAULT_INSTANCE).updateCharacteristicValue(inputReportCharacteristic.getValueHandle(), (uint8_t*)consumerEmptyInputReportData, sizeof(consumerInputReportData));
        HID_DEBUG("result \%d \r\n", result);
    }
    HID_DEBUG("button_pressed ended");
}


int main(){
   BLE& ble = BLE::Instance(BLE::DEFAULT_INSTANCE);
   ble.init(bleInitComplete);
   bt.mode(PullUp);
   bt.rise(button_pressed);
    
    while(true){
        ble.waitForEvent();
    
    }    
}